/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.blocks.devices;

import java.util.function.Predicate;
import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.block.BlockHorizontal;
import net.minecraft.block.SoundType;
import net.minecraft.block.material.Material;
import net.minecraft.block.state.BlockFaceShape;
import net.minecraft.block.state.BlockStateContainer;
import net.minecraft.block.state.IBlockState;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.player.EntityPlayer;
import net.minecraft.init.Blocks;
import net.minecraft.init.SoundEvents;
import net.minecraft.item.ItemStack;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumHand;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.math.AxisAlignedBB;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.RayTraceResult;
import net.minecraft.util.math.Vec3d;
import net.minecraft.world.IBlockAccess;
import net.minecraft.world.World;

import net.dries007.tfc.api.capability.size.IItemSize;
import net.dries007.tfc.api.capability.size.Size;
import net.dries007.tfc.api.capability.size.Weight;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.property.ILightableBlock;
import net.dries007.tfc.objects.items.ItemFireStarter;
import net.dries007.tfc.objects.te.TEBloomery;
import net.dries007.tfc.util.Helpers;
import net.dries007.tfc.util.Multiblock;

import static net.minecraft.block.BlockTrapDoor.OPEN;

@ParametersAreNonnullByDefault
public class BlockBloomery extends BlockHorizontal implements IItemSize, ILightableBlock
{
    //[horizontal index][basic shape / door1 / door2]
    private static final AxisAlignedBB[][] AABB =
        {
            {
                new AxisAlignedBB(0.0F, 0.0F, 0.0f, 1.0f, 1.0F, 0.5F),
                new AxisAlignedBB(0.0F, 0.0F, 0.0f, 0.125f, 1.0F, 0.5F),
                new AxisAlignedBB(0.875F, 0.0F, 0.0f, 1.0f, 1.0F, 0.5F)
            },
            {
                new AxisAlignedBB(0.5F, 0.0F, 0.0F, 1.0F, 1.0F, 1.0f),
                new AxisAlignedBB(0.5F, 0.0F, 0.0F, 1.0F, 1.0F, 0.125f),
                new AxisAlignedBB(0.5F, 0.0F, 0.875F, 1.0F, 1.0F, 1.0f)
            },
            {
                new AxisAlignedBB(0.0f, 0.0F, 0.5F, 1.0F, 1.0F, 1.0F),
                new AxisAlignedBB(0.0f, 0.0F, 0.5F, 0.125F, 1.0F, 1.0F),
                new AxisAlignedBB(0.875f, 0.0F, 0.5F, 1.0F, 1.0F, 1.0F)
            },
            {
                new AxisAlignedBB(0.0F, 0.0F, 0.0F, 0.5F, 1.0F, 1.0F),
                new AxisAlignedBB(0.0F, 0.0F, 0.0F, 0.5F, 1.0F, 0.125F),
                new AxisAlignedBB(0.0F, 0.0F, 0.875F, 0.5F, 1.0F, 1.0F)
            }
        };

    private static final Multiblock BLOOMERY_CHIMNEY, BLOOMERY_BASE, GATE_NORTH, GATE_SOUTH, GATE_EAST, GATE_WEST;

    static
    {
        Predicate<IBlockState> stoneMatcher = state -> state.func_185904_a() == Material.field_151576_e && state.func_185915_l();
        Predicate<IBlockState> insideChimney = state -> state.func_177230_c() == BlocksTFC.MOLTEN || state.func_177230_c() == Blocks.field_150350_a;
        Predicate<IBlockState> center = state -> state.func_177230_c() == BlocksTFC.CHARCOAL_PILE || state.func_177230_c() == BlocksTFC.BLOOM || state.func_177230_c() == Blocks.field_150350_a;
        BLOOMERY_BASE = new Multiblock()
            .match(new BlockPos(0, 0, 0), center)
            .match(new BlockPos(0, -1, 0), stoneMatcher);
        BLOOMERY_CHIMNEY = new Multiblock()
            .match(new BlockPos(0, 0, 0), insideChimney)
            .match(new BlockPos(1, 0, 0), stoneMatcher)
            .match(new BlockPos(-1, 0, 0), stoneMatcher)
            .match(new BlockPos(0, 0, 1), stoneMatcher)
            .match(new BlockPos(0, 0, -1), stoneMatcher);
        //Only one of the gates will return true if the structure is built right
        GATE_NORTH = new Multiblock()
            .match(new BlockPos(0, 0, 1), state -> state.func_177230_c() == BlocksTFC.BLOOMERY || state.func_177230_c() == Blocks.field_150350_a)
            .match(new BlockPos(1, 0, 1), stoneMatcher)
            .match(new BlockPos(-1, 0, 1), stoneMatcher)
            .match(new BlockPos(0, 1, 1), stoneMatcher)
            .match(new BlockPos(0, -1, 1), stoneMatcher)
            .match(new BlockPos(0, 0, -1), stoneMatcher)
            .match(new BlockPos(1, 0, 0), stoneMatcher)
            .match(new BlockPos(-1, 0, 0), stoneMatcher);
        GATE_SOUTH = new Multiblock()
            .match(new BlockPos(0, 0, -1), state -> state.func_177230_c() == BlocksTFC.BLOOMERY || state.func_177230_c() == Blocks.field_150350_a)
            .match(new BlockPos(1, 0, -1), stoneMatcher)
            .match(new BlockPos(-1, 0, -1), stoneMatcher)
            .match(new BlockPos(0, 1, -1), stoneMatcher)
            .match(new BlockPos(0, -1, -1), stoneMatcher)
            .match(new BlockPos(0, 0, 1), stoneMatcher)
            .match(new BlockPos(1, 0, 0), stoneMatcher)
            .match(new BlockPos(-1, 0, 0), stoneMatcher);
        GATE_WEST = new Multiblock()
            .match(new BlockPos(1, 0, 0), state -> state.func_177230_c() == BlocksTFC.BLOOMERY || state.func_177230_c() == Blocks.field_150350_a)
            .match(new BlockPos(1, 0, -1), stoneMatcher)
            .match(new BlockPos(1, 0, 1), stoneMatcher)
            .match(new BlockPos(1, 1, 0), stoneMatcher)
            .match(new BlockPos(1, -1, 0), stoneMatcher)
            .match(new BlockPos(0, 0, 1), stoneMatcher)
            .match(new BlockPos(0, 0, -1), stoneMatcher)
            .match(new BlockPos(-1, 0, 0), stoneMatcher);
        GATE_EAST = new Multiblock()
            .match(new BlockPos(-1, 0, 0), state -> state.func_177230_c() == BlocksTFC.BLOOMERY || state.func_177230_c() == Blocks.field_150350_a)
            .match(new BlockPos(-1, 0, -1), stoneMatcher)
            .match(new BlockPos(-1, 0, 1), stoneMatcher)
            .match(new BlockPos(-1, 1, 0), stoneMatcher)
            .match(new BlockPos(-1, -1, 0), stoneMatcher)
            .match(new BlockPos(0, 0, 1), stoneMatcher)
            .match(new BlockPos(0, 0, -1), stoneMatcher)
            .match(new BlockPos(1, 0, 0), stoneMatcher);
    }

    public BlockBloomery()
    {
        super(Material.field_151573_f);
        func_149672_a(SoundType.field_185852_e);
        setHarvestLevel("pickaxe", 0);
        func_149711_c(20.0F);
        func_180632_j(this.field_176227_L.func_177621_b()
            .func_177226_a(field_185512_D, EnumFacing.NORTH)
            .func_177226_a(LIT, false)
            .func_177226_a(field_176283_b, false));
    }

    public int getChimneyLevels(World world, BlockPos centerPos)
    {
        for (int i = 1; i < 4; i++)
        {
            BlockPos center = centerPos.func_177981_b(i);
            if (!BLOOMERY_CHIMNEY.test(world, center))
            {
                return i - 1;
            }
        }
        // Maximum levels
        return 3;
    }

    public boolean isFormed(World world, BlockPos centerPos, EnumFacing facing)
    {
        if (!BLOOMERY_BASE.test(world, centerPos)) return false;
        if (getChimneyLevels(world, centerPos) == 0) return false;
        switch (facing)
        {
            case NORTH:
                return GATE_NORTH.test(world, centerPos);
            case SOUTH:
                return GATE_SOUTH.test(world, centerPos);
            case EAST:
                return GATE_EAST.test(world, centerPos);
            case WEST:
                return GATE_WEST.test(world, centerPos);
        }
        return false;
    }

    @Override
    @Nonnull
    public Size getSize(ItemStack stack)
    {
        return Size.VERY_SMALL;
    }

    @Override
    @Nonnull
    public Weight getWeight(ItemStack stack)
    {
        return Weight.HEAVY;
    }

    @Override
    @SuppressWarnings("deprecation")
    @Nonnull
    public IBlockState func_176203_a(int meta)
    {
        return this.func_176223_P()
            .func_177226_a(field_185512_D, EnumFacing.func_176731_b(meta % 4))
            .func_177226_a(LIT, meta / 4 % 2 != 0)
            .func_177226_a(field_176283_b, meta / 8 != 0);
    }

    @Override
    public int func_176201_c(IBlockState state)
    {
        return state.func_177229_b(field_185512_D).func_176736_b()
            + (state.func_177229_b(LIT) ? 4 : 0)
            + (state.func_177229_b(field_176283_b) ? 8 : 0);
    }

    @Override
    @SuppressWarnings("deprecation")
    public boolean func_149686_d(IBlockState state)
    {
        return false;
    }

    @Override
    @SuppressWarnings("deprecation")
    @Nonnull
    public AxisAlignedBB func_185496_a(IBlockState state, IBlockAccess source, BlockPos pos)
    {
        return AABB[state.func_177229_b(field_185512_D).func_176736_b()][0];
    }

    @Override
    @SuppressWarnings("deprecation")
    @Nonnull
    public BlockFaceShape func_193383_a(IBlockAccess worldIn, IBlockState state, BlockPos pos, EnumFacing face)
    {
        return BlockFaceShape.UNDEFINED;
    }

    @Override
    @SuppressWarnings("deprecation")
    @ParametersAreNonnullByDefault
    public AxisAlignedBB func_180646_a(IBlockState blockState, IBlockAccess worldIn, BlockPos pos)
    {
        if (blockState.func_177229_b(field_176283_b))
        {
            return field_185506_k;
        }
        return AABB[blockState.func_177229_b(field_185512_D).func_176736_b()][0];
    }

    @Override
    @SuppressWarnings("deprecation")
    public boolean func_149662_c(IBlockState state)
    {
        return false;
    }

    @Override
    public void func_180663_b(World worldIn, BlockPos pos, IBlockState state)
    {
        TEBloomery te = Helpers.getTE(worldIn, pos, TEBloomery.class);
        if (te != null)
        {
            te.onBreakBlock(worldIn, pos);
        }
        super.func_180663_b(worldIn, pos, state);
    }

    @Override
    @SuppressWarnings("deprecation")
    @Nullable
    @ParametersAreNonnullByDefault
    public RayTraceResult func_180636_a(IBlockState blockState, World worldIn, BlockPos pos, Vec3d start, Vec3d end)
    {
        if (blockState.func_177229_b(field_176283_b))
        {
            int index = blockState.func_177229_b(field_185512_D).func_176736_b();
            RayTraceResult rayTraceDoor1 = func_185503_a(pos, start, end, AABB[index][1]), rayTraceDoor2 = func_185503_a(pos, start, end, AABB[index][2]);

            if (rayTraceDoor1 == null)
            {
                return rayTraceDoor2;
            }
            else if (rayTraceDoor2 == null)
            {
                return rayTraceDoor1;
            }
            if (rayTraceDoor1.field_72307_f.func_72436_e(end) > rayTraceDoor2.field_72307_f.func_72436_e(end))
            {
                return rayTraceDoor1;
            }
            return rayTraceDoor2;
        }
        return super.func_180636_a(blockState, worldIn, pos, start, end);
    }

    @Override
    public boolean func_176196_c(World worldIn, @Nonnull BlockPos pos)
    {
        return super.func_176196_c(worldIn, pos) && getAValidFacing(worldIn, pos) != null;
    }

    @Override
    public boolean func_180639_a(World worldIn, BlockPos pos, IBlockState state, EntityPlayer player, EnumHand hand, EnumFacing facing, float hitX, float hitY, float hitZ)
    {
        if (!worldIn.field_72995_K)
        {
            if (!state.func_177229_b(LIT))
            {
                worldIn.func_175656_a(pos, state.func_177231_a(field_176283_b));
                worldIn.func_184133_a(null, pos, SoundEvents.field_187610_bh, SoundCategory.BLOCKS, 1.0f, 1.0f);
            }
            TEBloomery te = Helpers.getTE(worldIn, pos, TEBloomery.class);
            if (te != null)
            {
                if (!state.func_177229_b(LIT) && te.canIgnite())
                {
                    ItemStack held = player.func_184586_b(hand);
                    if (ItemFireStarter.canIgnite(held))
                    {
                        worldIn.func_175656_a(pos, state.func_177226_a(LIT, true).func_177226_a(field_176283_b, false));
                        te.onIgnite();
                        return true;
                    }
                }
            }
        }
        return true;
    }

    @Override
    @SuppressWarnings("deprecation")
    @Nonnull
    public IBlockState func_180642_a(World worldIn, BlockPos pos, EnumFacing facing, float hitX, float hitY, float hitZ, int meta, EntityLivingBase placer)
    {
        EnumFacing placeDirection = getAValidFacing(worldIn, pos);
        if (placeDirection != null)
        {
            return this.func_176223_P().func_177226_a(field_185512_D, placeDirection);
        }
        return super.func_180642_a(worldIn, pos, facing, hitX, hitY, hitZ, meta, placer);
    }

    @Override
    @Nonnull
    protected BlockStateContainer func_180661_e()
    {
        return new BlockStateContainer(this, field_185512_D, LIT, field_176283_b);
    }

    @Override
    public int getLightValue(IBlockState state, IBlockAccess world, BlockPos pos)
    {
        return state.func_177229_b(LIT) ? 15 : 0;
    }

    @Override
    public boolean hasTileEntity(IBlockState state)
    {
        return true;
    }

    @Override
    public TileEntity createTileEntity(World world, IBlockState state)
    {
        return new TEBloomery();
    }

    @Nullable
    private EnumFacing getAValidFacing(World world, BlockPos pos)
    {
        for (EnumFacing facing : EnumFacing.field_176754_o)
        {
            if (isFormed(world, pos.func_177972_a(facing), facing))
            {
                return facing;
            }
        }
        return null;
    }
}
