/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.objects.te;

import java.util.HashSet;
import java.util.Set;
import javax.annotation.ParametersAreNonnullByDefault;

import net.minecraft.block.Block;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.EnumParticleTypes;
import net.minecraft.util.SoundCategory;
import net.minecraft.util.math.BlockPos;
import net.minecraft.util.math.Vec3i;
import net.minecraftforge.fml.relauncher.Side;
import net.minecraftforge.fml.relauncher.SideOnly;

import net.dries007.tfc.TerraFirmaCraft;
import net.dries007.tfc.api.util.IBellowsConsumerBlock;
import net.dries007.tfc.client.TFCSounds;
import net.dries007.tfc.network.PacketBellowsUpdate;
import net.dries007.tfc.objects.blocks.devices.BlockCharcoalForge;
import net.dries007.tfc.objects.blocks.devices.BlockFirePit;

import static net.minecraft.block.BlockHorizontal.FACING;

@ParametersAreNonnullByDefault
public class TEBellows extends TEBase
{
    public static final Vec3i OFFSET_LEVEL = new Vec3i(1, 0, 0);
    public static final Vec3i OFFSET_INSET = new Vec3i(1, -1, 0);

    private static final Set<Vec3i> offsets = new HashSet<>();
    private static final int BELLOWS_AIR = 200;

    static
    {
        addBellowsOffset(OFFSET_LEVEL);
        addBellowsOffset(OFFSET_INSET);
    }

    /**
     * Notify the bellows that it should check a certain offset when blowing air
     * X: front of the bellows, positive values go forward
     * Y: vertical, self-explanatory, negative is below.
     * Z: X but rotated 90 degrees clockwise, positive values go right. in most cases you want this to be 0
     * For example: a block that must sit right in front of the bellows(like fire pits) must check {@code Vec3i(1,0,0)},
     * {@link BlockFirePit}
     * meanwhile, blocks that sink into ground( like forges) must check {@code Vec3i(1,-1,0)}.
     * {@link BlockCharcoalForge}
     *
     * @param offset The offset to check
     */
    public static void addBellowsOffset(Vec3i offset)
    {
        offsets.add(offset);
    }

    private long lastPushed = 0L;

    // Min 0.125, Max 0.625
    @SideOnly(Side.CLIENT)
    public double getHeight()
    {
        int time = (int) (field_145850_b.func_82737_E() - lastPushed);
        if (time < 10)
            return (double) time * 0.05 + 0.125;
        else if (time < 20)
            return (double) (20 - time) * 0.05 + 0.125;
        return 0.125;
    }

    public void onReceivePacket(long lastPushed)
    {
        this.lastPushed = lastPushed;
    }

    public boolean onRightClick()
    {
        long time = field_145850_b.func_82737_E() - lastPushed;
        if (time < 20)
            return true;
        field_145850_b.func_184133_a(null, field_174879_c, TFCSounds.BELLOWS_BLOW_AIR, SoundCategory.BLOCKS, 1.0F, 1.0F);

        if (!field_145850_b.field_72995_K)
        {
            lastPushed = field_145850_b.func_82737_E();
            TerraFirmaCraft.getNetwork().sendToDimension(new PacketBellowsUpdate(this, lastPushed), field_145850_b.field_73011_w.getDimension());
        }

        EnumFacing direction = field_145850_b.func_180495_p(field_174879_c).func_177229_b(field_185512_D); // It is a better idea to inherit the direction directly from the block.
        for (Vec3i offset : offsets)
        {
            BlockPos posx = field_174879_c.func_177981_b(offset.func_177956_o())
                .func_177967_a(direction, offset.func_177958_n())
                .func_177967_a(direction.func_176746_e(), offset.func_177952_p());
            Block block = field_145850_b.func_180495_p(posx).func_177230_c();
            if (block instanceof IBellowsConsumerBlock && ((IBellowsConsumerBlock) block).canIntakeFrom(this, offset, direction))
            {
                ((IBellowsConsumerBlock) block).onAirIntake(this, field_145850_b, posx, BELLOWS_AIR);
                if (field_145850_b.field_72995_K)
                {
                    //TODO: actual sound, better particles and animation
                    // old bellows just shot particles outwards, that was simpler.
                    posx = field_174879_c.func_177972_a(direction);
                    field_145850_b.func_175688_a(EnumParticleTypes.SMOKE_LARGE, posx.func_177958_n() + .5d, posx.func_177956_o() + .5d, posx.func_177952_p() + .5d, 0, 0, 0);
                }
                return true;
            }
        }
        return true;
    }

    public void debug()
    {
        TerraFirmaCraft.getLog().debug("Debugging Bellows");
        TerraFirmaCraft.getLog().debug("Now: {} | Then: {} | Difference: {}", field_145850_b.func_82737_E(), lastPushed, (field_145850_b.func_82737_E() - lastPushed));
        TerraFirmaCraft.getLog().debug("Total Height: {}", getHeight());
    }
}
