/*
 * Work under Copyright. Licensed under the EUPL.
 * See the project README.md and LICENSE.txt for more information.
 */

package net.dries007.tfc.world.classic.worldgen;

import java.util.Random;

import net.minecraft.block.state.IBlockState;
import net.minecraft.util.math.BlockPos;
import net.minecraft.world.EnumSkyBlock;
import net.minecraft.world.World;
import net.minecraft.world.chunk.IChunkProvider;
import net.minecraft.world.gen.IChunkGenerator;
import net.minecraftforge.fml.common.IWorldGenerator;

import net.dries007.tfc.ConfigTFC;
import net.dries007.tfc.api.registries.TFCRegistries;
import net.dries007.tfc.api.types.Plant;
import net.dries007.tfc.api.types.Rock;
import net.dries007.tfc.objects.blocks.BlocksTFC;
import net.dries007.tfc.objects.blocks.plants.BlockPlantTFC;
import net.dries007.tfc.objects.blocks.stone.BlockRockVariant;
import net.dries007.tfc.world.classic.ChunkGenTFC;
import net.dries007.tfc.world.classic.ClimateTFC;
import net.dries007.tfc.world.classic.WorldTypeTFC;
import net.dries007.tfc.world.classic.chunkdata.ChunkDataTFC;

/**
 * todo: make these bigger without causing cascading lag.
 * This will require larger re-writes on the scale of oregen
 * Wait for 1.14+ as AlcatrazEscapee is doing a worldgen rewrite anyway
 */
public class WorldGenSoilPits implements IWorldGenerator
{
    private static final float CLAY_RAINFALL_THREHOLD = 100f;

    @Override
    public void generate(Random random, int chunkX, int chunkZ, World world, IChunkGenerator chunkGenerator, IChunkProvider chunkProvider)
    {
        if (!(chunkGenerator instanceof ChunkGenTFC)) return;
        final BlockPos chunkBlockPos = new BlockPos(chunkX << 4, 0, chunkZ << 4);

        BlockPos pos = world.func_175672_r(chunkBlockPos.func_177982_a(8 + random.nextInt(16), 0, 8 + random.nextInt(16)));
        generateClay(world, random, pos);

        pos = world.func_175672_r(chunkBlockPos.func_177982_a(8 + random.nextInt(16), 0, 8 + random.nextInt(16)));
        generatePeat(world, random, pos);
    }

    private void generateClay(World world, Random rng, BlockPos start)
    {
        // If this has to have a radius that is >= 8, then it needs to be moved to a cascading-lag safe model
        // Otherwise, do not change this unless you are prepared to do some fairly large re-writes, similar to how ore gen is handled
        int radius = rng.nextInt(6) + 2;
        int depth = rng.nextInt(3) + 1;
        if (rng.nextInt(ConfigTFC.WORLD.clayRarity) != 0 || start.func_177956_o() > WorldTypeTFC.SEALEVEL + 6) return;
        if (ChunkDataTFC.getRainfall(world, start) < CLAY_RAINFALL_THREHOLD) return;

        for (int x = -radius; x <= radius; x++)
        {
            for (int z = -radius; z <= radius; z++)
            {
                if (x * x + z * z > radius * radius) continue;
                final BlockPos posHorizontal = start.func_177982_a(x, 0, z);

                boolean flag = false;
                for (int y = -depth; y <= +depth; y++)
                {
                    final BlockPos pos = posHorizontal.func_177982_a(0, y, 0);
                    final IBlockState current = world.func_180495_p(pos);
                    if (BlocksTFC.isDirt(current))
                    {
                        world.func_180501_a(pos, BlockRockVariant.get(ChunkDataTFC.getRockHeight(world, pos), Rock.Type.CLAY).func_176223_P(), 2);
                        flag = true;
                    }
                    else if (BlocksTFC.isGrass(current))
                    {
                        world.func_180501_a(pos, BlockRockVariant.get(ChunkDataTFC.getRockHeight(world, pos), Rock.Type.CLAY_GRASS).func_176223_P(), 2);
                        flag = true;
                    }
                }
                if (flag && rng.nextInt(15) == 0)
                {
                    final BlockPos pos = world.func_175672_r(posHorizontal);

                    for (Plant plant : TFCRegistries.PLANTS.getValuesCollection())
                    {
                        if (plant.getIsClayMarking())
                        {
                            BlockPlantTFC plantBlock = BlockPlantTFC.get(plant);
                            IBlockState state = plantBlock.func_176223_P();
                            int plantAge = plant.getAgeForWorldgen(rng, ClimateTFC.getHeightAdjustedTemp(world, pos));

                            if (!world.field_73011_w.func_177495_o() && !world.func_189509_E(pos) &&
                                plant.isValidLocation(ClimateTFC.getHeightAdjustedTemp(world, pos), ChunkDataTFC.getRainfall(world, pos), world.func_175642_b(EnumSkyBlock.SKY, pos)) &&
                                world.func_175623_d(pos) &&
                                plantBlock.func_180671_f(world, pos, state))
                            {
                                world.func_180501_a(pos, state.func_177226_a(BlockPlantTFC.AGE, plantAge), 2);
                            }
                        }
                    }
                }
            }
        }
    }

    private boolean generatePeat(World world, Random rng, BlockPos start)
    {
        // If this has to have a radius that is >= 8, then it needs to be moved to a cascading-lag safe model
        // Otherwise, do not change this unless you are prepared to do some fairly large re-writes, similar to how ore gen is handled
        int radius = rng.nextInt(4) + 4;
        byte depth = 2;
        boolean flag = false;

        if (rng.nextInt(30) != 0 || start.func_177956_o() > WorldTypeTFC.SEALEVEL) return false;

        for (int x = -radius; x <= radius; ++x)
        {
            for (int z = -radius; z <= radius; ++z)
            {
                if (x * x + z * z > radius * radius) continue;

                for (int y = -depth; y <= depth; ++y)
                {
                    final BlockPos pos = start.func_177982_a(x, y, z);
                    if (!ClimateTFC.isSwamp(world, pos)) continue;
                    final IBlockState current = world.func_180495_p(pos);

                    if (BlocksTFC.isGrass(current))
                    {
                        world.func_180501_a(pos, BlocksTFC.PEAT_GRASS.func_176223_P(), 2);
                        flag = true;
                    }
                    else if (BlocksTFC.isDirt(current) || BlocksTFC.isClay(current))
                    {
                        world.func_180501_a(pos, BlocksTFC.PEAT.func_176223_P(), 2);
                        flag = true;
                    }
                }
            }
        }
        return flag;
    }
}
